#Script contact: 
#Brenna Schwert
#Remote Sensing Specialist
#RedCastle Resources, Inc.
#Forest Service Contractor
#Remote Sensing Applications Center (RSAC)
#801-975-3821
#bmschwert@fs.fed.us
###############################################################################################################################################################
#
#  Input parameters.  Nothing else needs to be edited.
#
###############################################################################################################################################################

#Folder where files will be written to
workspace = ""

#Point shapefile that serves as the "training data"
pointsshp = ""

#The attribute name of the field in the point shapefile you want to predict
classfield = "Class"

#A value of "Thematic" or "Continuous".  Whether or not you want to do classification or regression
predicttype = "Thematic"

#A list of all the continuous predictor images (e.g. Landsat)
imageList = c("")

#A list of all the thematic predictor images (e.g. soils).  This is commonly empty (i.e. thematic imagelist = c(""))
thematicimagelist = c("")

#The output image location and name.
OutputModel = ""
###############################################################################################################################################################

#install and get required packages
install.packages("raster", repos='http://cran.us.r-project.org')
install.packages("randomForest", repos='http://cran.us.r-project.org')
install.packages("rgdal", repos='http://cran.us.r-project.org')
install.packages("tools", repos='http://cran.us.r-project.org')

library(raster)
library(randomForest)
library(tools)
library(rgdal)

#set working directory
setwd(workspace)

#read in points
points = shapefile(pointsshp)

rm(pointsshp)
gc()
rasterOptions(chunksize = 2e+06)
######################################################################################
#
#  create stacklist
#
######################################################################################

stacklist = c()

#read each raster from imageList as stack
print("Stacking Rasters")
if (length(imageList) > 0)
{
	for(i in 1:length(imageList))
	{
		rast = stack(imageList[i])
		stacklist = append(stacklist,rast)
		rm(rast)
	}
}

if (length(thematicimagelist) > 0 & thematicimagelist[1] != "")
{
	for(i in 1:length(thematicimagelist))
	{
		rast = stack(thematicimagelist[i])
		stacklist = append(stacklist,rast)
		rm(rast)
	}
}

print(stacklist)

#stack all the rasters
ourStack = stack(stacklist)

#plot rasters
plotRGB(ourStack, r=4, g=3,b=2,stretch = 'hist')

rm(stacklist)
gc()

######################################################################################
#
# extract points
#
######################################################################################

print("Extracting Point Values")
pointvalues = extract(ourStack, points)
pointsDF = as.data.frame(points)
head(pointsDF)
viewPts=pointsDF
coordinates(viewPts) <- ~x+y #defining the co-ordiantes in the data frame
head(viewPts)

#plot(viewPts, pch=16, col="yellow", add=TRUE) #plot all the points
#points(viewPts, pch=16, col= pointsDF$Class_Num) #plot by class

rm(points)
gc()

######################################################################################
#
# get class field
#
######################################################################################

classindex = which(colnames(pointsDF)==classfield)

rm(classfield)
gc()

######################################################################################
#
#  Recode the thematic values
#
######################################################################################

legend = ""
emptyvec = c()

if(predicttype=="Thematic")
{
	classnames = as.factor(pointsDF[,classindex])
	numvec = seq(1,length(levels(classnames)),1)
	legend = as.data.frame(cbind(numvec,levels(classnames)))
	
	#loop through and change values in pointsDF
	for(i in 1:dim(pointsDF)[1])
	{
		newvalue = as.numeric(as.character(legend[which(legend[,2]==pointsDF[i,classindex]),][1]))
		emptyvec = append(emptyvec,newvalue)
	}
	
	colnames(legend) = c("NumValue","TextValue")
	
	rm(newvalue)
	rm(numvec)
	rm(classnames)
	gc()

}
	
######################################################################################
#
# create modeldataset
#
######################################################################################

print("Creating Model Dataset")

if(predicttype=="Thematic"){
	ModelDataset = as.data.frame(cbind(emptyvec, pointvalues))
}else{
	ModelDataset = as.data.frame(cbind(pointsDF[,classindex], pointvalues))
}

#if it should be a thematic output, then force "class" field to a factor
if(predicttype=="Thematic")
{
	ModelDataset[,1] = as.factor(ModelDataset[,1])
}

rm(emptyvec)	
rm(classindex)
rm(pointvalues)
rm(pointsDF)
gc()

######################################################################################
#  
#  set data type for the rasters
#
######################################################################################

#set continuous rasters to numeric
if (length(imageList)>0)
{
	print("Setting Rasters to continuous")
	for( i in 1:length(imageList))
	{
		ImageName = basename(file_path_sans_ext(imageList[i]))
		for(j in 1:dim(ModelDataset)[2])
		{
			columnname = unlist(strsplit(basename(colnames(ModelDataset)[j]),split = "\\."))[1] 
			
			if(ImageName == columnname)
			{
				ModelDataset[,j] = as.numeric(ModelDataset[,j])	
				print("Changed following column to numeric:")
				print(colnames(ModelDataset)[j])
				print(is.numeric(ModelDataset[,j]))
			}
		}
		rm(columnname)
		rm(ImageName)
		gc()
	}
}

#set thematic rasters to factor	
if(length(thematicimagelist)>0)
{
	print("Setting Rasters to factor")
	for( i in 1:length(thematicimagelist))
	{
		ImageName = basename(file_path_sans_ext(thematicimagelist[i]))
		for(j in 1:dim(ModelDataset)[2])
		{	
			columnname = basename(colnames(ModelDataset)[j])
			if(ImageName == columnname)
			{
				ModelDataset[,j] = as.factor(ModelDataset[,j])	
				print("Changed following column to factor:")
				print(colnames(ModelDataset)[j])
				print(is.factor(ModelDataset[,j]))
			}
		}
		rm(columnname)
		rm(ImageName)
		gc()
	}
}

rm(imageList)
rm(thematicimagelist)
gc()

######################################################################################

colnames(ModelDataset)[1] = "Class"

######################################################################################
#
# create random forest model
#
######################################################################################

print("Creating Random Forest Model")
RF_Model = randomForest(Class~., data=ModelDataset,importance=TRUE,ntree=1000, na.action=na.omit)
print(RF_Model)

rm(ModelDataset)
gc()

######################################################################################
#
# output to file varimpplot and confusion matrix
#
######################################################################################

base = basename(file_path_sans_ext(OutputModel))

#write the legend
if(predicttype=="Thematic")
{
	output_confusion = paste(base,"_confusion_matrix.txt",sep="",collapse = "")
	
	confusion_matrix = as.data.frame(RF_Model[5])
	rfType=as.data.frame(RF_Model[2])
	errorRate=mean(RF_Model[4]$err.rate[,1])
	test=cbind(rfType, errorRate)
	write.table(test, output_confusion)
	write.table(confusion_matrix, output_confusion, append = TRUE)
	legendout = paste(base,"_legend.csv",sep="",collapse = "")
	write.csv(legend, legendout,row.names=FALSE)
	rm(legendout)
	rm(confusion_matrix)
	rm(output_confusion)
	gc()
}

varimp = paste(base,"_VarImpPlot.jpg",sep="",collapse = "")
png(filename=varimp)
varImpPlot(RF_Model, type = 1)
dev.off()

rm(varimp)
rm(legend)
rm(base)
gc()

######################################################################################
#
# create predict raster
#
######################################################################################

time1 = Sys.time()

print("Predicting Raster")
if(predicttype=="Thematic"){
	outputrast = predict(ourStack, RF_Model, filename = OutputModel, type='response',progress = "text", datatype = 'INT1U', inf.rm = TRUE, overwrite = TRUE)
}else{
	outputrast = predict(ourStack, RF_Model, filename = OutputModel, type='response',progress = "text", datatype = 'INT1U', inf.rm = TRUE, overwrite = TRUE)
}

#PLOTTING THE PREDICTED RASTER in R#
plot(outputrast)

time2 = Sys.time()
totaltime = time2 - time1
print(totaltime)

rm(outputrast)
rm(RF_Model)
rm(predicttype)
rm(ourStack)
rm(OutputModel)
rm(time1)
rm(time2)
rm(totaltime)
gc()
